@@ -31,7 +31,7 @@ def main(app_config=None, x=0):
 
     with alice:
         # Create EPR pair with Bob, through the repeater.
-        epr = epr_socket.create()[0]
+        epr = epr_socket.create_keep()[0]
 
         # CHSH strategy: measure in one of 2 bases depending on x.
         if x == 0:
