@@ -16,8 +16,8 @@ from itertools import count
 from typing import (
     TYPE_CHECKING,
     Callable,
+    ContextManager,
     Dict,
-    Iterator,
     List,
     Optional,
     Tuple,
@@ -37,6 +37,12 @@ from netqasm.backend.messages import (
 from netqasm.lang import operand
 from netqasm.lang.ir import BreakpointAction, BreakpointRole, PreSubroutine
 from netqasm.logging.glob import get_netqasm_logger
+from netqasm.sdk.build_types import (
+    GenericHardwareConfig,
+    HardwareConfig,
+    T_BranchRoutine,
+    T_LoopRoutine,
+)
 from netqasm.sdk.compiling import SubroutineCompiler
 from netqasm.sdk.config import LogConfig
 from netqasm.sdk.futures import Array, Future, RegFuture, T_CValue
@@ -46,22 +52,16 @@ from netqasm.sdk.qubit import Qubit
 from netqasm.sdk.shared_memory import SharedMemory, SharedMemoryManager
 from netqasm.util.log import LineTracker
 
-from .builder import Builder
+from .builder import Builder, SdkLoopUntilContext
 
 # Generic type for messages sent to the quantum node controller.
 # Note that `SubroutineMessage` does not derive from `Message` so it has to be
 # mentioned explicitly.
 T_Message = Union[Message, SubroutineMessage]
 
-# Callback function type for conditional statements.
-T_BranchRoutine = Callable[["BaseNetQASMConnection"], None]
-
-# Callback function type for loop statements.
-T_LoopRoutine = Callable[["BaseNetQASMConnection"], None]
-
 # Imports that are only needed for type checking
 if TYPE_CHECKING:
-    from netqasm.sdk.epr_socket import EPRSocket
+    from netqasm.sdk import epr_socket as esck
 
 
 class BaseNetQASMConnection(abc.ABC):
@@ -94,8 +94,9 @@ class BaseNetQASMConnection(abc.ABC):
         node_name: Optional[str] = None,
         app_id: Optional[int] = None,
         max_qubits: int = 5,
+        hardware_config: Optional[HardwareConfig] = None,
         log_config: LogConfig = None,
-        epr_sockets: Optional[List[EPRSocket]] = None,
+        epr_sockets: Optional[List[esck.EPRSocket]] = None,
         compiler: Optional[Type[SubroutineCompiler]] = None,
         return_arrays: bool = True,
         _init_app: bool = True,
@@ -130,6 +131,10 @@ class BaseNetQASMConnection(abc.ABC):
             at the same time by applications registered through this connection.
             Defaults to 5.
 
+        :param hardware_config: configuration object with info about the underlying
+            hardware. Used by the Builder of this Connection. When None,
+            a generic configuration object is created with the default qubit count of 5.
+
         :param log_config: configuration object specifying what to log.
 
         :param epr_sockets: list of EPR sockets.
@@ -189,13 +194,16 @@ class BaseNetQASMConnection(abc.ABC):
         # Set directory for storing serialized subroutines.
         self._log_subroutines_dir: Optional[str] = log_config.log_subroutines_dir
 
+        if hardware_config is None:
+            hardware_config = GenericHardwareConfig(max_qubits)
+
         # Builder object of this connection.
         # The Builder is used to gather application code and output an
         # Intermediate Representation (IR), which is then compiled into subroutines.
         self._builder = Builder(
             connection=self,
             app_id=self._app_id,
-            max_qubits=max_qubits,
+            hardware_config=hardware_config,
             compiler=compiler,
             return_arrays=return_arrays,
         )
@@ -246,6 +254,10 @@ class BaseNetQASMConnection(abc.ABC):
     def network_info(self) -> Type[NetworkInfo]:
         return self._get_network_info()
 
+    @property
+    def builder(self) -> Builder:
+        return self._builder
+
     @classmethod
     def get_app_ids(cls) -> Dict[str, List[int]]:
         return cls._app_ids
@@ -410,7 +422,7 @@ class BaseNetQASMConnection(abc.ABC):
 
         :return: list of active qubits
         """
-        return self._builder.active_qubits
+        return self._builder._mem_mgr.get_active_qubits()
 
     def _init_new_app(self, max_qubits: int) -> None:
         """Send a message to the quantum node controller to register a new application.
@@ -432,7 +444,7 @@ class BaseNetQASMConnection(abc.ABC):
         while self.shared_memory is None:
             pass
 
-    def _setup_epr_sockets(self, epr_sockets: Optional[List[EPRSocket]]) -> None:
+    def _setup_epr_sockets(self, epr_sockets: Optional[List[esck.EPRSocket]]) -> None:
         """Send messages to the quantum node controller to open EPR sockets."""
         if epr_sockets is None:
             return
@@ -478,7 +490,7 @@ class BaseNetQASMConnection(abc.ABC):
         :param callback: if `block` is False, this callback is called when the quantum
             node controller sends the subroutine results.
         """
-        subroutine = self._builder._pop_pending_subroutine()
+        subroutine = self._builder.subrt_pop_pending_subroutine()
         if subroutine is None:
             return
 
@@ -503,7 +515,7 @@ class BaseNetQASMConnection(abc.ABC):
         self._logger.debug(f"Flushing presubroutine:\n{presubroutine}")
 
         # Parse, assembly and possibly compile the subroutine
-        subroutine = self._builder._pre_process_subroutine(presubroutine)
+        subroutine = self._builder.subrt_compile_subroutine(presubroutine)
         self._logger.info(f"Flushing compiled subroutine:\n{subroutine}")
 
         # Commit the subroutine to the quantum device
@@ -538,7 +550,7 @@ class BaseNetQASMConnection(abc.ABC):
             have the same length as `length`.
         :return: a handle to the array that can be used in application code
         """
-        return self._builder.new_array(length, init_values)
+        return self._builder.alloc_array(length, init_values)
 
     def loop(
         self,
@@ -546,7 +558,7 @@ class BaseNetQASMConnection(abc.ABC):
         start: int = 0,
         step: int = 1,
         loop_register: Optional[operand.Register] = None,
-    ) -> Iterator[operand.Register]:
+    ) -> ContextManager[operand.Register]:
         """Create a context for code that gets looped.
 
         Each iteration of the loop is associated with an index, which starts at 0
@@ -563,6 +575,7 @@ class BaseNetQASMConnection(abc.ABC):
         (compiled) code that is inside the context.
 
         Example:
+
         .. code-block::
 
             with NetQASMConnection(app_name="alice") as alice:
@@ -580,9 +593,7 @@ class BaseNetQASMConnection(abc.ABC):
             In most cases there is no need to explicitly specify this.
         :return: the context object (to be used in a `with ...` expression)
         """
-        # TODO: this returns a method that has a decorator.
-        #       Are type hints still correct?
-        return self._builder.loop(stop, start, step, loop_register)  # type: ignore
+        return self._builder.sdk_loop_context(stop, start, step, loop_register)
 
     def loop_body(
         self,
@@ -590,7 +601,7 @@ class BaseNetQASMConnection(abc.ABC):
         stop: int,
         start: int = 0,
         step: int = 1,
-        loop_register: Optional[operand.Register] = None,
+        loop_register: Optional[Union[operand.Register, str]] = None,
     ) -> None:
         """Loop code that is defined in a Python function (body).
 
@@ -603,7 +614,32 @@ class BaseNetQASMConnection(abc.ABC):
         :param step: step size of iteration range, defaults to 1
         :param loop_register: specific register to be used for holding the loop index.
         """
-        self._builder.loop_body(body, stop, start, step, loop_register)
+        self._builder.sdk_loop_body(body, stop, start, step, loop_register)
+
+    def loop_until(self, max_iterations: int) -> ContextManager[SdkLoopUntilContext]:
+        """Create a context with code to be looped until the exit condition is met, or
+        the maximum number of tries has been reached.
+
+        The code inside the context is automatically looped (re-run).
+        At the end of each iteration, the exit_condition of the context object is
+        checked. If the condition holds, the loop exits. Otherwise the loop
+        does another iteration. If `max_iterations` iterations have been reached,
+        the loop exits anyway.
+
+        Make sure you set the loop_condition on the context object, like e.g.
+
+        .. code-block::
+
+        with connection.loop_until(max_iterations=10) as loop:
+            q = Qubit(conn)
+            m = q.measure()
+            constraint = ValueAtMostConstraint(m, 0)
+            loop.set_exit_condition(constraint)
+
+
+        :param max_iterations: the maximum number of times to loop
+        """
+        return self.builder.sdk_new_loop_until_context(max_iterations)
 
     def if_eq(self, a: T_CValue, b: T_CValue, body: T_BranchRoutine) -> None:
         """Execute a function if a == b.
@@ -617,7 +653,7 @@ class BaseNetQASMConnection(abc.ABC):
         :param b: a classical value
         :param body: function to execute if condition is true
         """
-        self._builder.if_eq(a, b, body)
+        self._builder.sdk_if_eq(a, b, body)
 
     def if_ne(self, a: T_CValue, b: T_CValue, body: T_BranchRoutine) -> None:
         """Execute a function if a != b.
@@ -631,7 +667,7 @@ class BaseNetQASMConnection(abc.ABC):
         :param b: a classical value
         :param body: function to execute if condition is true
         """
-        self._builder.if_ne(a, b, body)
+        self._builder.sdk_if_ne(a, b, body)
 
     def if_lt(self, a: T_CValue, b: T_CValue, body: T_BranchRoutine) -> None:
         """Execute a function if a < b.
@@ -645,7 +681,7 @@ class BaseNetQASMConnection(abc.ABC):
         :param b: a classical value
         :param body: function to execute if condition is true
         """
-        self._builder.if_lt(a, b, body)
+        self._builder.sdk_if_lt(a, b, body)
 
     def if_ge(self, a: T_CValue, b: T_CValue, body: T_BranchRoutine) -> None:
         """Execute a function if a > b.
@@ -659,7 +695,7 @@ class BaseNetQASMConnection(abc.ABC):
         :param b: a classical value
         :param body: function to execute if condition is true
         """
-        self._builder.if_ge(a, b, body)
+        self._builder.sdk_if_ge(a, b, body)
 
     def if_ez(self, a: T_CValue, body: T_BranchRoutine) -> None:
         """Execute a function if a == 0.
@@ -672,7 +708,7 @@ class BaseNetQASMConnection(abc.ABC):
         :param a: a classical value
         :param body: function to execute if condition is true
         """
-        self._builder.if_ez(a, body)
+        self._builder.sdk_if_ez(a, body)
 
     def if_nz(self, a: T_CValue, body: T_BranchRoutine) -> None:
         """Execute a function if a != 0.
@@ -685,7 +721,11 @@ class BaseNetQASMConnection(abc.ABC):
         :param a: a classical value
         :param body: function to execute if condition is true
         """
-        self._builder.if_nz(a, body)
+        self._builder.sdk_if_nz(a, body)
+
+    def try_until_success(self, max_tries: int = 1) -> ContextManager[None]:
+        """TODO docstring"""
+        return self._builder.sdk_try_context(max_tries)
 
     def tomography(
         self,
@@ -788,7 +828,7 @@ class BaseNetQASMConnection(abc.ABC):
     def insert_breakpoint(
         self, action: BreakpointAction, role: BreakpointRole = BreakpointRole.CREATE
     ) -> None:
-        self._builder.insert_breakpoint(action, role)
+        self._builder._build_cmds_breakpoint(action, role)
 
 
 class DebugConnection(BaseNetQASMConnection):
