@@ -11,13 +11,15 @@ from typing import TYPE_CHECKING, List, Optional, Union
 
 from netqasm.lang import operand
 from netqasm.lang.ir import GenericInstr, ICmd, Symbols
+from netqasm.lang.operand import Address, ArrayEntry
 from netqasm.lang.parsing import parse_address, parse_register
 from netqasm.typedefs import T_Cmd
 from netqasm.util.log import HostLine
 
 if TYPE_CHECKING:
-    from netqasm.lang.ir import T_OperandUnion
-    from netqasm.sdk.connection import BaseNetQASMConnection
+    from netqasm.lang import ir
+    from netqasm.sdk import connection as sdkconn
+    from netqasm.sdk.builder import Builder, SdkForEachContext, SdkIfContext
 
 # Generic type for classical values (that may only get a value at runtime).
 T_CValue = Union[int, "Future", "RegFuture"]
@@ -146,13 +148,17 @@ class BaseFuture(int):
     def __new__(cls, *args, **kwargs):
         return int.__new__(cls, 0)
 
-    def __init__(self, connection: BaseNetQASMConnection):
+    def __init__(self, connection: sdkconn.BaseNetQASMConnection):
         self._value: Optional[int] = None
-        self._connection: BaseNetQASMConnection = connection
+        self._connection: sdkconn.BaseNetQASMConnection = connection
 
     def __repr__(self):
         return f"{self.__class__} with value={self.value}"
 
+    @property
+    def builder(self) -> Builder:
+        return self._connection.builder
+
     @property
     def value(self) -> Optional[int]:
         """Get the value of the future.
@@ -184,54 +190,38 @@ class BaseFuture(int):
         :param other: value to add to this Future's value
         :param mod: do the addition modulo `mod`
         """
-        raise NotImplementedError(f"add is not implement for {self.__class__.__name__}")
-
-    def if_eq(self, other: Optional[T_CValue]) -> _IfContext:
-        return _IfContext(
-            connection=self._connection,
-            condition=GenericInstr.BEQ,
-            a=self,
-            b=other,
+        raise NotImplementedError(
+            f"add is not implemented for {self.__class__.__name__}"
         )
 
-    def if_ne(self, other: Optional[T_CValue]) -> _IfContext:
-        return _IfContext(
-            connection=self._connection,
-            condition=GenericInstr.BNE,
-            a=self,
-            b=other,
+    def if_eq(self, other: Optional[T_CValue]) -> SdkIfContext:
+        return self.builder.sdk_new_if_context(
+            condition=GenericInstr.BEQ, op0=self, op1=other
         )
 
-    def if_lt(self, other: Optional[T_CValue]) -> _IfContext:
-        return _IfContext(
-            connection=self._connection,
-            condition=GenericInstr.BLT,
-            a=self,
-            b=other,
+    def if_ne(self, other: Optional[T_CValue]) -> SdkIfContext:
+        return self.builder.sdk_new_if_context(
+            condition=GenericInstr.BNE, op0=self, op1=other
         )
 
-    def if_ge(self, other: Optional[T_CValue]) -> _IfContext:
-        return _IfContext(
-            connection=self._connection,
-            condition=GenericInstr.BGE,
-            a=self,
-            b=other,
+    def if_lt(self, other: Optional[T_CValue]) -> SdkIfContext:
+        return self.builder.sdk_new_if_context(
+            condition=GenericInstr.BLT, op0=self, op1=other
         )
 
-    def if_ez(self) -> _IfContext:
-        return _IfContext(
-            connection=self._connection,
-            condition=GenericInstr.BEZ,
-            a=self,
-            b=None,
+    def if_ge(self, other: Optional[T_CValue]) -> SdkIfContext:
+        return self.builder.sdk_new_if_context(
+            condition=GenericInstr.BGE, op0=self, op1=other
         )
 
-    def if_nz(self) -> _IfContext:
-        return _IfContext(
-            connection=self._connection,
-            condition=GenericInstr.BNZ,
-            a=self,
-            b=None,
+    def if_ez(self) -> SdkIfContext:
+        return self.builder.sdk_new_if_context(
+            condition=GenericInstr.BEZ, op0=self, op1=None
+        )
+
+    def if_nz(self) -> SdkIfContext:
+        return self.builder.sdk_new_if_context(
+            condition=GenericInstr.BNZ, op0=self, op1=None
         )
 
 
@@ -247,9 +237,9 @@ class Future(BaseFuture):
 
     def __init__(
         self,
-        connection: BaseNetQASMConnection,
+        connection: sdkconn.BaseNetQASMConnection,
         address: int,
-        index: Union[int, Future, operand.Register],
+        index: Union[int, Future, operand.Register, RegFuture],
     ):
         """Future constructor. Typically not used directly.
 
@@ -260,7 +250,7 @@ class Future(BaseFuture):
         """
         super().__init__(connection=connection)
         self._address: int = address
-        self._index: Union[int, Future, operand.Register] = index
+        self._index: Union[int, Future, operand.Register, RegFuture] = index
 
     def __str__(self) -> str:
         value = self.value
@@ -296,8 +286,8 @@ class Future(BaseFuture):
             other = parse_register(other)
 
         # Store self in a temporary register
-        tmp_register = self._connection._builder._get_inactive_register(activate=True)
-        load_commands = self._get_load_commands(tmp_register)
+        tmp_register = self.builder._mem_mgr.get_inactive_register(activate=True)
+        load_commands = self.get_load_commands(tmp_register)
         store_commands = self._get_store_commands(tmp_register)
 
         other_tmp_register: Optional[operand.Register] = None
@@ -305,18 +295,16 @@ class Future(BaseFuture):
 
         # If other is a Future, also load this into a temporary register
         if isinstance(other, Future):
-            other_operand = self._connection._builder._get_inactive_register(
-                activate=True
-            )
+            other_operand = self.builder._mem_mgr.get_inactive_register(activate=True)
             other_tmp_register = other_operand
-            load_commands += other._get_load_commands(other_tmp_register)
+            load_commands += other.get_load_commands(other_tmp_register)
             store_commands += other._get_store_commands(other_tmp_register)
         elif isinstance(other, operand.Register) or isinstance(other, int):
             other_operand = other
         else:
             raise NotImplementedError("for type {type(other)}")
 
-        add_operands: List[T_OperandUnion] = [
+        add_operands: List[ir.T_OperandUnion] = [
             tmp_register,
             tmp_register,
             other_operand,
@@ -340,13 +328,15 @@ class Future(BaseFuture):
             + store_commands
         )
 
-        self._connection._builder._remove_active_register(tmp_register)
+        self.builder._mem_mgr.remove_active_register(tmp_register)
         if other_tmp_register is not None:
-            self._connection._builder._remove_active_register(other_tmp_register)
+            self.builder._mem_mgr.remove_active_register(other_tmp_register)
 
-        self._connection._builder.add_pending_commands(commands)
+        self.builder.subrt_add_pending_commands(commands)
 
-    def _get_load_commands(self, register: operand.Register) -> List[T_Cmd]:
+    def get_load_commands(self, register: operand.Register) -> List[T_Cmd]:
+        """Return a list of PreSubroutine commands for loading this Future into
+        the specified register."""
         return self._get_access_commands(GenericInstr.LOAD, register)
 
     def _get_store_commands(self, register: operand.Register) -> List[T_Cmd]:
@@ -364,15 +354,20 @@ class Future(BaseFuture):
                 raise RuntimeError(
                     "Future-index must be from the same connection as the future itself"
                 )
-            tmp_register = self._connection._builder._get_inactive_register()
+            tmp_register = self.builder._mem_mgr.get_inactive_register()
             # NOTE this might be many commands if the index is a future with a future index etc
-            with self._connection._builder._activate_register(tmp_register):
+            with self.builder._activate_register(tmp_register):
                 access_index_cmds = self._index._get_access_commands(
                     instruction=GenericInstr.LOAD,
                     register=tmp_register,
                 )
             commands += access_index_cmds
             index: Union[int, operand.Register] = tmp_register
+        elif isinstance(self._index, RegFuture):
+            assert (
+                self._index.reg is not None
+            ), "Trying to use RegFuture that has no value yet"
+            index = self._index.reg
         elif isinstance(self._index, int) or isinstance(self._index, operand.Register):
             index = self._index
         else:
@@ -392,6 +387,20 @@ class Future(BaseFuture):
         commands.append(access_cmd)
         return commands
 
+    def get_address_entry(self) -> ArrayEntry:
+        """Convert this Future to an ArrayEntry object to be used an instruction
+        operand."""
+        if isinstance(self._index, RegFuture):
+            assert self._index.reg is not None, (
+                f"cannot use RegFuture {self._index} as array index since "
+                f"it does not yet have a value"
+            )
+            return ArrayEntry(Address(self._address), self._index.reg)
+        elif isinstance(self._index, int) or isinstance(self._index, operand.Register):
+            return ArrayEntry(Address(self._address), self._index)
+        else:
+            assert False, f"index type {self._index} not supported"
+
 
 class RegFuture(BaseFuture):
     """Represents a single register value that will become available in the future.
@@ -400,7 +409,9 @@ class RegFuture(BaseFuture):
     """
 
     def __init__(
-        self, connection: BaseNetQASMConnection, reg: Optional[operand.Register] = None
+        self,
+        connection: sdkconn.BaseNetQASMConnection,
+        reg: Optional[operand.Register] = None,
     ):
         """RegFuture constructor. Typically not used directly.
 
@@ -459,18 +470,16 @@ class RegFuture(BaseFuture):
 
         # If other is a Future, also load this into a temporary register
         if isinstance(other, Future):
-            other_operand = self._connection._builder._get_inactive_register(
-                activate=True
-            )
+            other_operand = self.builder._mem_mgr.get_inactive_register(activate=True)
             other_tmp_register = other_operand
-            load_commands += other._get_load_commands(other_tmp_register)
+            load_commands += other.get_load_commands(other_tmp_register)
             store_commands += other._get_store_commands(other_tmp_register)
         elif isinstance(other, operand.Register) or isinstance(other, int):
             other_operand = other
         else:
             raise NotImplementedError("for type {type(other)}")
 
-        add_operands: List[T_OperandUnion] = [
+        add_operands: List[ir.T_OperandUnion] = [
             self.reg,
             self.reg,
             other_operand,
@@ -495,9 +504,9 @@ class RegFuture(BaseFuture):
         )
 
         if other_tmp_register is not None:
-            self._connection._builder._remove_active_register(other_tmp_register)
+            self.builder._mem_mgr.remove_active_register(other_tmp_register)
 
-        self._connection._builder.add_pending_commands(commands)
+        self.builder.subrt_add_pending_commands(commands)
 
 
 class Array:
@@ -512,7 +521,7 @@ class Array:
 
     def __init__(
         self,
-        connection: BaseNetQASMConnection,
+        connection: sdkconn.BaseNetQASMConnection,
         length: int,
         address: int,
         init_values: Optional[List[Optional[int]]] = None,
@@ -531,7 +540,7 @@ class Array:
                 raise TypeError("Array needs to consist of int's or None's")
             length = len(init_values)
         assert isinstance(length, int) and length > 0, f"{length} is not a valid length"
-        self._connection: BaseNetQASMConnection = connection
+        self._connection: sdkconn.BaseNetQASMConnection = connection
         self._length: int = length
         self._address: int = address
         self._init_values: Optional[List[Optional[int]]] = init_values
@@ -556,7 +565,13 @@ class Array:
     def address(self) -> int:
         return self._address
 
-    def get_future_index(self, index: Union[int, str, operand.Register]) -> Future:
+    @property
+    def builder(self) -> Builder:
+        return self._connection.builder
+
+    def get_future_index(
+        self, index: Union[int, str, operand.Register, RegFuture]
+    ) -> Future:
         """Get a Future representing a particular array element"""
         if isinstance(index, str):
             index = parse_register(index)
@@ -581,7 +596,7 @@ class Array:
                 range_args.append(x)
         return [self.get_future_index(index) for index in range(*range_args)]
 
-    def foreach(self) -> _ForEachContext:
+    def foreach(self) -> SdkForEachContext:
         """Create a context of code that gets called for each element in the array.
 
         Returns a future of the array value at the current index.
@@ -592,6 +607,7 @@ class Array:
         No classical communication is allowed.
 
         Example:
+
         .. code-block::
 
             with NetQASMConnection(app_name="alice") as alice:
@@ -606,13 +622,9 @@ class Array:
                         q.H()
                     q.measure(future=outcomes.get_future_index(i))
         """
-        return _ForEachContext(
-            connection=self._connection,
-            array=self,
-            return_index=False,
-        )
+        return self.builder.sdk_new_foreach_context(array=self, return_index=False)
 
-    def enumerate(self) -> _ForEachContext:
+    def enumerate(self) -> SdkForEachContext:
         """Create a context of code that gets called for each element in the array
         and includes a counter.
 
@@ -625,6 +637,7 @@ class Array:
         No classical communication is allowed.
 
         Example:
+
         .. code-block::
 
             with NetQASMConnection(app_name="alice") as alice:
@@ -639,79 +652,8 @@ class Array:
                         q.H()
                     q.measure(future=outcomes.get_future_index(i))
         """
-        return _ForEachContext(
-            connection=self._connection,
-            array=self,
-            return_index=True,
-        )
+        return self.builder.sdk_new_foreach_context(array=self, return_index=True)
 
-
-class _Context:
-
-    next_id: int = 0
-
-    @property
-    @abc.abstractmethod
-    def ENTER_METH(self):
-        pass
-
-    @property
-    @abc.abstractmethod
-    def EXIT_METH(self):
-        pass
-
-    def __init__(self, connection: BaseNetQASMConnection, **kwargs):
-        self._id: int = self._get_id()
-        self._connection: BaseNetQASMConnection = connection
-        self._kwargs = kwargs
-
-    def _get_id(self) -> int:
-        _Context.next_id += 1
-        return _Context.next_id - 1
-
-    def __enter__(self):
-        return getattr(self._connection._builder, self.ENTER_METH)(
-            context_id=self._id,
-            **self._kwargs,
-        )
-
-    def __exit__(self, *args, **kwargs):
-        getattr(self._connection._builder, self.EXIT_METH)(
-            context_id=self._id,
-            **self._kwargs,
-        )
-
-
-class _IfContext(_Context):
-
-    ENTER_METH = "_enter_if_context"
-    EXIT_METH = "_exit_if_context"
-
-    def __init__(
-        self,
-        connection: BaseNetQASMConnection,
-        condition: GenericInstr,
-        a: Optional[T_CValue],
-        b: Optional[T_CValue],
-    ):
-        super().__init__(
-            connection=connection,
-            condition=condition,
-            a=a,
-            b=b,
-        )
-
-
-class _ForEachContext(_Context):
-
-    ENTER_METH = "_enter_foreach_context"
-    EXIT_METH = "_exit_foreach_context"
-
-    def __init__(
-        self, connection: BaseNetQASMConnection, array: Array, return_index: bool
-    ):
-        super().__init__(
-            connection=connection,
-            array=array,
-            return_index=return_index,
-        )
+    def undefine(self) -> None:
+        """Undefine (i.e. set to 'None') all elements in the array."""
+        self.builder._build_cmds_undefine_array(self)
