@@ -37,3 +37,7 @@ def test_enumerate():
         ]
     )
     simulate_application(app_instance, use_app_config=False, enable_logging=False)
+
+
+if __name__ == "__main__":
+    test_enumerate()
